import java.io.*;

/**
 * Main class that runs the arithmetically encodes and decodes an image.
 * 
 * Usage:
 * java BitPlaneCoder inputImage encodedFile outputImage bpp
 * 
 * where
 * inputImage = the image to encode
 * encodedFile = the file to write the encoded result to (will also be read from for the decoding process)
 * outputImage = the file to write the decoded image to
 * bpp = the desired stopping bitrate (bits/pixel)
 *
 * @author Amanda Askew & Dane Barney
 * @version 2003.10.01
 */
public class BitPlaneCoder {
    private static final String USAGE = "Usage:\n"+
                                        "java BitPlaneCoder inputImage encodedFile outputImage bpp [contextfile]";

    private static final int BLOCKSIZE = 8;

    public static void main(String[] args) {
        if(args.length != 4) {
            System.out.println("Invalid syntax. "+USAGE);
            return;
        }
        
        try {
            String context = "";
            // interpret command line args
            String input = args[0];
            String encoded = args[1];
            String output = args[2];
            double bpp = Double.valueOf(args[3]).doubleValue();
            if (args.length > 4)
              context = args[4];
            if(bpp < 0) throw new InvalidFormatException("Bits/pixel cannot be negative.");
            
            // read image from file
            Image im = new Image(input, true);
            System.out.println("Input: "+input+"\n"+
                               "\t     type: "+im.sayType()+"\n"+
                               "\t    width: "+im.width+"\n"+
                               "\t   height: "+im.height+"\n"+
                               "\tbit-depth: "+im.bitDepth+"\n"+
                               "\t filesize: "+im.filesize+" bytes\n");
            

	    DiscreteCosineTransform transform = new DiscreteCosineTransform(BLOCKSIZE, im.width, im.height);


            /**************************************
             *       Begin Encoding Process       *
             **************************************/
            
            // perform forward wavelet transform
            System.out.print("Performing discrete cosine transform...");
            transform.forwardTransform(im);
            System.out.println("done.");
            
            // subtract average value from DC subband coefficients
            transform.subtractDCavg();
            
            // find the smallest power of 2 greater than the largest abs val coefficient
            transform.getUpperBound();
            
            // create encoder object, which will handle all writing to encoded file
            BitPlaneEncoder enc = new BitPlaneEncoder(im, transform, encoded,context);
            
            // encode transform coefficients to file
            System.out.println("Encoding...");
            enc.encode(bpp);
            
            // print results
            int bef = im.filesize;
            int aft = enc.out.bytesWritten;
            System.out.println("Original filesize: "+bef+" bytes\nCompressed filesize: "+aft+" bytes");
            System.out.println("Compression ratio: "+(double)bef/aft+":1, "+(double)(bef-aft)/bef*100+"%");
            System.out.println("Compression rate: "+(double)enc.encodedBytes*8/(im.width*im.height)+" bits/pixel (Originally "+im.bitDepth+" bits/pixel)");
            System.out.println();
            
            // free memory
            transform = new DiscreteCosineTransform(BLOCKSIZE, im.width, im.height);
            enc = null;
            // could instead use the same image header and transform for decoder,
            // completely bypassing the need for an intermediate encoded file


            /**************************************
             *       Begin Decoding Process       *
             **************************************/
            
            // create decoder object, which will handle all reading from encoded file
            BitPlaneDecoder dec = new BitPlaneDecoder(encoded, transform);
            
            // decode transform coefficients from file
            System.out.println("Decoding...");
            transform = dec.decode();

            // add average value back to DC subband coefficients
            transform.addDCavg();
            
            // make a copy of the original image's header info
            Image im2 = new Image(im, true);
            
            // perform inverse wavelet transform to get image pixels back
            System.out.print("Performing inverse transform...");
            im2.pixels = transform.inverseTransform();
	    System.out.println("done.");
 	    System.out.println("im2.width: " + im2.width);
	    System.out.println("im2.height: " + im2.height);
           
            // compute psnr of the two images
            Tools.printPSNR(im, im2);
            
            // save image to file
            System.out.print("Writing output file "+output+"...");
            im2.save(output);
            System.out.println("done.");
   
        } catch(Throwable t) {
            System.out.println(t);
            t.printStackTrace(System.out);
        }
    }

}
