import java.io.*;

/**
 * The class which performs the decoding process.
 * 
 * @author Amanda Askew & Dane Barney
 * @version 2003.10.01
 */
public class BitPlaneDecoder extends BitPlaneEndec {
    private BitReader in;

    /**
     * Opens the compressedFile for reading and interprets all information in the file
     * header (before the encoded data), and initializes the output stream, arithmetic coder, 
     * wavelet transform, contexts, and priority queue.
     * 
     * @param compressedFile the file to decode
     */
    public BitPlaneDecoder(String compressedFile, DiscreteCosineTransform transform) throws InvalidFormatException, IOException {
        in = new BitReader(new DataInputStream(new FileInputStream(compressedFile)));
        ar = new ArithmeticCoder();

        this.transform = transform;

        isSig = new boolean[transform.nSubbands][];
        signPlane = new byte[transform.nSubbands][];
        for(int subband=0; subband < transform.nSubbands; subband++) {
            isSig[subband] = new boolean[transform.subbandSize];
            signPlane[subband] = new byte[transform.subbandSize];
        }
        
        // strip image header from file
        readImgHeader();
        
        readContextCoords();
        
        // create all the possible contexts and store them in an array
        createContextArray();

        // read upper bound
        transform.upperBound = in.readDouble();
	System.out.println("transform.upperBound: " + transform.upperBound);
        
        // read the average DC coeff value
        transform.DCavg = in.readDouble();
	System.out.println("transform.DCavg: " + transform.DCavg);
    }
    
    /**
     * Reads and interprets the image header data, which is stored in the compressed
     * file before the actual encoded data.
     */
    public void readImgHeader() throws InvalidFormatException {
        try {
            int type = in.readByte(true);
            im = new Image(type, in, true);
        } catch(IOException e) {
            // if an EOFException is thrown, the header is not properly formatted
            throw new InvalidFormatException("Invalid encoded file. Header is not recognized.");
        }
    }
     /**
     * Reads the positions of the pixels in the context that was used to encode this file.
     */
    public void readContextCoords() throws InvalidFormatException {
        try {
            contextCoords = new Coords[in.readByte(true)]; // first byte = total number of context coordinates
            for(int i=0; i < contextCoords.length; i++) {
                int x = in.readByte(true); // 1st byte = x coord
                int y = in.readByte(true); // 2nd byte = y coord
                contextCoords[i] = new Coords(x,y);
            }
        } catch(IOException e) {
            throw new InvalidFormatException("Invalid encoded file. Header is not recognized.");
        }
    }

    /**
     * Decodes transform coefficients from the encoded data and returns the 
     * resulting transform.
     */
    public DiscreteCosineTransform decode() throws IOException {
	
        // prepare for decoding
        ar.initializeTag(in);
	DoublyLinkedList refinList = new DoublyLinkedList();
        DoublyLinkedList newlySigCoeffs = new DoublyLinkedList();
        double upperBound = transform.upperBound; // duplicate for later use

        int bitPlane;
 	
	/* TODO: The next two for loops read in the signPlane for all coefficients
	 * and entire successive bit planes until we are out of input.  
	 * You need to make this decoder match your encoder.
	 *
	 */
        for (int subband =0; subband<transform.nSubbands; subband++) {
              for (int coeff=0; coeff<transform.subbandSize; coeff++) {
		  signPlane[subband][coeff] = (byte)ar.decodeValue(signContext, in);
		  // stop if end of file
		  if(in.eof && ar.doneDecoding()) break;
	      }
	      if(in.eof && ar.doneDecoding()) break;
	}

	

        for(bitPlane = 1; bitPlane <= Tools.DOUBLE_MANTISSA_BITS; bitPlane++) {

            System.out.print("\tbit plane "+bitPlane+" ");

            // prepare to write coefficients for this plane
            currentPlane = new byte[transform.nSubbands][];

            // reset which bits have already been coded for this plane
            for(int subband=0; subband < transform.nSubbands; subband++)
                currentPlane[subband] = new byte[transform.subbandSize];

            // reset the list of newly significant coefficients for this plane
            newlySigCoeffs = new DoublyLinkedList();

            // while there are still bits to be encoded
	    for (int subband =0; subband<transform.nSubbands; subband++) {
		for (int coeff=0; coeff<transform.subbandSize; coeff++) {
		    
		    // decode bit and put in current bit plane
		    Context context = findContext(subband,coeff);
		    currentPlane[subband][coeff] = (byte)ar.decodeValue(context, in);
		    // stop if end of file
		  if(in.eof && ar.doneDecoding()) break;
                }
		if(in.eof && ar.doneDecoding()) break;
	    }
	    
	    
            // append current plane to subbandPtr
            addPlane(currentPlane);
            
            // stop if end of file
            if(in.eof && ar.doneDecoding()) break;
            
            
            System.out.println("finished");

        } // end main for loop
        
        // report if last bit plane was not finished
        if(in.eof && ar.doneDecoding()) System.out.println("incomplete; end of bitstream");

        // add sign back onto unsigned coefficients
        addSign();

        in.close();
        
        return transform;
    }
    
    private void addPlane(byte[][] plane) {
        transform.upperBound /= 2;
        for(int subband=0; subband < transform.nSubbands; subband++) {
            for(int coeff=0; coeff < transform.subbandSize; coeff++) {
                byte bit = plane[subband][coeff];
                assert (bit == 0 || bit == 1);
                
                transform.subbandPtr[subband][coeff] += bit * transform.upperBound;
            }
        }
    }
    
    private void addSign() {
        for(int subband=0; subband < transform.nSubbands; subband++) {
            for(int coeff=0; coeff < transform.subbandSize; coeff++) {
                transform.subbandPtr[subband][coeff] *= signPlane[subband][coeff] == Tools.NEGATIVE ? -1 : 1;
            }
        }
    }
    
}
