import java.io.*;

/**
 * The class which performs the encoding process.
 * 
 * @author Amanda Askew & Dane Barney
 * @version 2003.09.11
 */
public class BitPlaneEncoder extends BitPlaneEndec {
    public BitWriter out;
    public int encodedBytes;
    private int startingBytes;
  
    /**
     * Initializes the output stream, arithmetic coder, image, wavelet transform,
     * contexts, and priority queue; and writes the file header (which precedes 
     * the actual encoded data).
     */
    public BitPlaneEncoder(Image inputImage, DiscreteCosineTransform transform, String encodedFile, String contextFile) throws InvalidFormatException, IOException  { 
        out = new BitWriter(new DataOutputStream(new FileOutputStream(encodedFile)));
        im = inputImage;
        ar = new ArithmeticCoder();

        this.transform = transform;

        isSig = new boolean[transform.nSubbands][];
        for(int subband=0; subband < transform.nSubbands; subband++)
            isSig[subband] = new boolean[transform.subbandSize];
        getSignPlane();
        
        // create all the possible contexts and store them in an array
        contextCoords = ContextReader.getContext(contextFile);
        createContextArray();
             
        // write image header to file
        writeImgHeader();

        // write context coordinates to file
        writeContextCoords();

        // write upperbound
        out.writeDouble(transform.upperBound);
        
        // write the average DC coeff value
        out.writeDouble(transform.DCavg);
        
    }
    
    /**
     * Writes the image header data, which is the first thing stored
     * in the compressed file.
     */
    public void writeImgHeader() throws IOException {
        out.writeByte((byte)im.type); // first send the type of image
        // now send the header byte-by-byte
        for(int i=0; i < im.header.length; i++)
            out.writeByte(im.header[i]);
    }
    
        /**
     * Writes the pixel coordinates in the context used to encode this file
     * (so the decoder knows).
     */
    public void writeContextCoords() throws IOException {
        out.writeByte((byte)contextCoords.length); // first write total number of context coordinates to write
        for(int i=0; i < contextCoords.length; i++) {
            out.writeByte((byte)contextCoords[i].x); // 1st byte = x coord
            out.writeByte((byte)contextCoords[i].y); // 2nd byte = y coord
        }
    }


    /**
     * Encodes the coefficients from the wavelet transform, writing to the output stream.
     */
    public void encode(double targetBpp) throws IOException {

        startingBytes = out.bytesWritten;
        
        int bitPlane;
        DoublyLinkedList refinList = new DoublyLinkedList();
        // main loop: iterate through bit planes until maximum precision is reached (defined by # of mantissa bits in the 64-bit IEEE-754 standard)


	/* TODO: The next two for loops in this method currently send all the signs (+/-) 
	 * and then all the bit planes up to target Bpp.  You need to replace them with 
	 * bit plane encoding (significance pass + refinement pass for each bit plane 
	 * up to target bpp).  Your decoder in BitPlaneDecoder.java should match.
	 *
	 */
	for (int subband =0; subband<transform.nSubbands; subband++) {
	    for (int coeff=0; coeff<transform.subbandSize; coeff++) {		 
		ar.encodeValue(signPlane[subband][coeff], signContext, out);
	    }
	    if(currentBpp() >= targetBpp) break;
	}
	
        for(bitPlane = 1; bitPlane < Tools.DOUBLE_MANTISSA_BITS; bitPlane++) {
            
	    System.gc();
            System.out.print("\tbit plane "+bitPlane+" ");

            // get the coefficient bits at this plane
            currentPlane = getNextPlane();

            // reset the list of newly significant coefficients for this plane
            DoublyLinkedList newlySigCoeffs = new DoublyLinkedList();

            // while there are still bits to be encoded
            for (int subband =0; subband<transform.nSubbands; subband++) {
              for (int coeff=0; coeff<transform.subbandSize; coeff++) {
		  int currentBit = currentPlane[subband][coeff];
		  Context context = findContext(subband,coeff);
		  ar.encodeValue(currentBit, context, out);
		  if(currentBpp() >= targetBpp) break;
	      }
	       if(currentBpp() >= targetBpp) break;
	    }
	    if(currentBpp() >= targetBpp) break;

	    System.out.println("finished");

        } // end main for loop

        // report if last bit plane was not finished
        if(currentBpp() >= targetBpp) System.out.println("incomplete; target bpp reached");

        ar.finishEncode(out);
        encodedBytes = out.bytesWritten - startingBytes;
        
        out.close();
    }
  
    /**
     * Strips off and returns the sign plane from the transform
     */
    private void getSignPlane() {
        signPlane = new byte[transform.nSubbands][];

        for(int subband=0; subband < transform.nSubbands; subband++) {
            signPlane[subband] = new byte[transform.subbandSize];
            
            for(int coeff=0; coeff < transform.subbandSize; coeff++) {
                double d = transform.subbandPtr[subband][coeff];
                
                // put a '0' or '1' into the sign plane, to represent pos or neg respectively
                signPlane[subband][coeff] = d > 0 ? Tools.POSITIVE : Tools.NEGATIVE;
                transform.subbandPtr[subband][coeff] = Math.abs(d);
            }
        }
    }
    
    /**
     * Strips off and returns the next plane from the transform, by modifying the coeffs subbandPtr
     * so they're ready for the next plane.
     */
    private byte[][] getNextPlane() {
        byte[][] bitPlane = new byte[transform.nSubbands][];

        for(int subband=0; subband < transform.nSubbands; subband++) {
            bitPlane[subband] = new byte[transform.subbandSize];

            for(int coeff=0; coeff < transform.subbandSize; coeff++) {
                double d = transform.subbandPtr[subband][coeff];

                // convert the unsigned value of the coefficient to a bit and scale it.
                // this conversion algorithm comes from CSE490gz lecture 5 slide 5
                if(d < transform.upperBound/2) {
                    bitPlane[subband][coeff] = Tools.INSIGNIFICANT;
                    transform.subbandPtr[subband][coeff] = 2*d;
                } else /* d >= transform.upperBound/2 */ {
                    bitPlane[subband][coeff] = Tools.SIGNIFICANT;
                    transform.subbandPtr[subband][coeff] = 2*d - transform.upperBound;
                }
            }
        }
        
        return bitPlane;
    }
    
    private double currentBpp() {
        double bitsWritten = 8.0*(out.bytesWritten+1 - startingBytes); // add 1 extra byte to account for "ar.finishEncode(out)"
        return bitsWritten / (im.width*im.height);
    }
    
}
