import java.io.*;

/**
 * The superclass which defines shared instance variables and methods used by
 * both BitPlaneEncoder and BitPlaneDecoder.
 * 
 * @author Amanda Askew & Dane Barney
 * @version 2003.09.11
 */
public class BitPlaneEndec {
    protected Image im;
    protected DiscreteCosineTransform transform;
    protected ArithmeticCoder ar;
    
    protected byte[][] signPlane;
    protected byte[][] currentPlane;

    protected boolean[][] isSig;

    protected Context[][] contexts; // 4 classes of contexts + 2
    /* |0|1|1|_|_|_|_|3|
     * |1|1|_|_|_|_|3|3|
     * |1|_|_|_|_|3|3|3|
     * |_|_|_|_|3|3|3|3|
     * |_|_|_|3|3|3|3|3|
     * |_|_|3|3|3|3|3|3|
     * |_|3|3|3|3|3|3|3|
     * |3|3|3|3|3|3|3|3|
     *
     * + sign context
     * + refinement context
     *
     *
     */

    protected Coords[] contextCoords;
    protected Context signContext; // special context for sign bits
    protected Context refinContext; // special context for refinement bits (after the 2nd one)
    
    /**
     * Creates an array of all the possible contexts as specified by their ordered pairs.
     * This consists of 2^((bit-depth)^(# of coordinates)) since each ordered pair can take on
     * 2^(bit-depth) values. Every context is initialized so that the count of all possible values
     * is equal (equal probabilities).
     */
    protected void createContextArray() {
	// get context coordinates from file
	System.out.print("Context coordinates: ");
	for(int i=0; i < contextCoords.length; i++) {
	    Coords cd = contextCoords[i];
	    System.out.print("("+cd.x+","+cd.y+") ");
	}
	System.out.println();

	int possibleValues = 2;
	int totalContexts = (int)Math.pow(possibleValues, contextCoords.length);
	contexts = new Context[4][totalContexts];
	for (int c = 0; c < 4; c+=1){
	    for(int i=0; i < totalContexts; i++) {
		int[] startCount = new int[possibleValues];
		// set start count for all possible values to 1
		for(int j=0; j < startCount.length; j++)
		    startCount[j] = 1;
		contexts[c][i] = new Context(startCount);
		contexts[c][i].indexNum = i;
	    }
	}
        int[] start1 = {1,1};
        int[] start2 = {1,1};
        signContext = new Context(start1);
        refinContext = new Context(start2);
    }

    /**
     * Finds the correct context to be applied to the coeff at (subband,coeff)
     */
    protected Context findContext(int subband, int coeff) {
        int hsize = transform.subbandHsize;
        int vsize = transform.subbandVsize;        
        int x = coeff % hsize;
        int y = coeff / hsize;
        int pixel;
        int contextNum =0;
        
	// first figure out which class of contexts we should be returning
	int context_class;
	if (subband == 0){
	    context_class = 0 ;
	}
	else if ((subband == 1) || (subband == 2)
		 || (subband == transform.blockSize)
		 || (subband == transform.blockSize + 1)
		 || (subband == transform.blockSize * 2)){
	    context_class = 1;
	}
	else {
	    int row = subband / transform.blockSize;
	    int col = subband % transform.blockSize;
	    if (row + col >= (transform.blockSize - 1)){
		context_class = 3;
	    }
	    else context_class = 2;
	}


        // iterate through all the coordinates defined in the context
        for(int k=0; k < contextCoords.length; k++) {
            
            int i = contextCoords[k].x; // neighbor's x
            int j = contextCoords[k].y; // neighbor's y
            int neighborCoeff = (j+y)*hsize + (i+x);
            
            // if one of the coordinates lies off the edge of the subband, assume it is white
            if ((i < 0 && i+x < 0) || // off the left edge
                (i > 0 && i+x > hsize-1) || // off the right edge
                (j < 0 && j+y < 0)) // off the top edge
		pixel = Tools.INSIGNIFICANT;

            else // normal case
		pixel = isSig[subband][neighborCoeff] ? 1 : 0;

            /* 
             * The context index number depends on the pixel value in each ordered pair position
             * and also depends on the bit-depth. The number of digits in the index number is
             * the number of ordered pairs. The pixel in each ordered pair position equals the 
             * digit at each position in the number. 2^(bit-depth) is the base of the number.
             * 
             * Ex. Suppose the bit-depth is 8 and the contexts consist of 3 ordered pairs. If one context
             * has pixel values 0, 7, and 3 in the first, second, and third ordered pairs respectively,
             * then the index number is 073 with a base of 2^8. This evaluates to:
             *   (0 * 256^2) + (7 * 256^1) + (3 * 256^0) = 1792
             */
            contextNum += pixel * (int)Math.pow(2, (contextCoords.length-1) - k);
        }
        return contexts[context_class][contextNum];
    }
  

}
