import java.io.*;

/**
 * Class which can read individual bits from an input stream.
 * 
 * @author Amanda Askew & Dane Barney
 * @version 2003.09.11
 */
public class BitReader {
    private DataInputStream in;
    private int shiftBits;
    private byte buffer;
    public boolean eof;
    
    public int num1s;
    public int num0s;

    /**
     * Initializes the input stream and sets shiftBits to its initial value of -1
     */
    public BitReader(DataInputStream in) {
        this.in = in;
        shiftBits = -1; // indicates that the buffer is empty
        eof = false;
    }
    
    public void close() throws IOException {
        in.close();
    }

    /**
     * Returns the next bit in the bitstream, then shifts the buffer over in preparation 
     * for the next read.
     * 
     * @param throwEOF  whether or not you want to throw an EOFException if the end 
     *                  of the inputstream has been reached. If false, just shifts
     *                  0's into the buffer.
     */
    public byte readBit(boolean throwEOF) throws IOException {
        if(shiftBits == -1) {
            try {
                buffer = in.readByte();                           
            } catch(EOFException e) {
                buffer = 0;
                eof = true;
            }
            shiftBits = 7;
        }
        if(eof && throwEOF)
            throw new EOFException();        
        byte temp = (byte)((buffer >>> shiftBits) & 1);
        shiftBits--;
        assert (shiftBits >= -1);
        assert (shiftBits < 7);
        if (temp == 0) num0s++;
        if (temp ==1) num1s++;
        return temp;
    }
    
    /**
     * Returns the next 4 bits in the bitstream, then shifts the buffer over in preparation 
     * for the next read.
     * 
     * @param throwEOF  whether or not you want to throw an EOFException if the end 
     *                  of the inputstream has been reached. If false, just shifts
     *                  0's into the buffer.
     */
    public byte readNibble(boolean throwEOF) throws IOException {
        byte temp = 0;
        for(int i=0; i < 4; i++) {
            temp <<= 1;
            temp |= readBit(throwEOF);
        }
        return temp;
    }

    /**
     * Returns the next 8 bits in the bitstream, then shifts the buffer over in preparation 
     * for the next read.
     * 
     * @param throwEOF  whether or not you want to throw an EOFException if the end 
     *                  of the inputstream has been reached. If false, just shifts
     *                  0's into the buffer.
     */
    public byte readByte(boolean throwEOF) throws IOException {
        byte temp = 0;
        if(shiftBits == -1) {
            try {
                temp = in.readByte();
            } catch(EOFException e) {
                temp = 0;
                eof = true;
            }
        } else {
            for(int i=0; i < 8; i++) {
                temp <<= 1;
                temp |= readBit(throwEOF);
            }
        }
        if(eof && throwEOF)
            throw new EOFException();
        return temp;
    }

    /**
     * Returns the next 16 bits in the bitstream, then shifts the buffer over in preparation 
     * for the next read.
     * 
     * @param throwEOF  whether or not you want to throw an EOFException if the end 
     *                  of the inputstream has been reached. If false, just shifts
     *                  0's into the buffer.
     */
    public short read2Bytes(boolean throwEOF) throws IOException {
        short temp = 0;
        if(shiftBits == -1) {
            try {
                temp = in.readShort();
            } catch(EOFException e) {
                temp = 0;
                eof = true;
            }
        } else {
            for(int i=0; i < 16; i++) {
                temp <<= 1;
                temp |= readBit(throwEOF);
            }
        }
        if(eof && throwEOF)
            throw new EOFException();
        return temp;
    }

    /**
     * Returns the next 24 bits in the bitstream, then shifts the buffer over in preparation 
     * for the next read.
     * 
     * @param throwEOF  whether or not you want to throw an EOFException if the end 
     *                  of the inputstream has been reached. If false, just shifts
     *                  0's into the buffer.
     */
    public int read3Bytes(boolean throwEOF) throws IOException {
        int temp = 0;
        if(shiftBits == -1) {
            try {
                temp = in.readShort() << 8;
                temp += in.readByte();
            } catch(EOFException e) {
                temp = 0;
                eof = true;
            }
        } else {
            for(int i=0; i < 24; i++) {
                temp <<= 1;
                temp |= readBit(throwEOF);
            }
        }
        if(eof && throwEOF)
            throw new EOFException();
        return temp;
    }
        
    /**
     * Returns the next 32 bits in the bitstream, then shifts the buffer over in preparation 
     * for the next read.
     * 
     * @param throwEOF  whether or not you want to throw an EOFException if the end 
     *                  of the inputstream has been reached. If false, just shifts
     *                  0's into the buffer.
     */
    public int read4Bytes(boolean throwEOF) throws IOException {
        int temp = 0;
        if(shiftBits == -1) {
            try {
                temp = in.readInt();
            } catch(EOFException e) {
                temp = 0;
                eof = true;
            }
        } else {
            for(int i=0; i < 32; i++) {
                temp <<= 1;
                temp |= readBit(throwEOF);
            }
        }
        if(eof && throwEOF)
            throw new EOFException();
        return temp;
    }

    public float readFloat() throws IOException {
        return in.readFloat();
    }
    
    public double readDouble() throws IOException {
        return in.readDouble();
    }
}
