import java.io.*;
import java.util.*;

/**
 * The class which reads and interprets the context layout from a text file.
 * 
 * The format of the text file is as follows:
 * A = the pixel in this location contributes to the context
 * X = ignore the pixel in this location
 * ? = the primary pixel
 * (NOTE: Use X's rather than spaces to the left of A's and ?)
 * 
 * Ex:
 * XXXA
 * AA
 * AXXX?
 * 
 * IMPORTANT: There cannot be any A's to the below and to the right of the ?,
 * because the pixels are coded in raster order.
 * 
 * @author Amanda Askew & Dane Barney
 * @version 2003.09.11
 */
public class ContextReader {
    private static final int MAX_SIZE = 31; // maximum number of neighbors involved in the context
    private static final int MAX_DISTANCE = 127; // maximum distance a neighbor can be from a pixel

    /**
     * Reads from the context file and converts the text matrix
     * into an array of ordered pairs which represent the (x,y) positions
     * of each neighbor, relative to the current pixel.
     * 
     * @param filename the text file being read from
     */
    public static Coords[] getContext(String filename) throws InvalidFormatException, IOException {

        Coords[] coords = new Coords[MAX_SIZE];

        try {
            BufferedReader inFile = new BufferedReader(new FileReader(filename));

            int yNum = 0;
            int arrayPos = 0;
            int qx = -1;
            int qy = -1;

            // read in one line at a time
            String line = inFile.readLine();

            while(line != null) {
                // iterate through all the characters in this line
                for(int i=0; i < line.length(); i++) {

                    char c = line.charAt(i);

                    if(c == 'A') { // 'A' corresponds to a pixel which will be used in the context
                        if(qx != -1) { // We require that all context pixels exist above and to the left of the '?' pixel.
                            inFile.close();
                            throw new InvalidFormatException("Invalid Context File: Found 'A' after '?'. The '?' symbol must be the last.");
                        }
                        try {
                            coords[arrayPos] = new Coords(i,yNum);
                            arrayPos++;
                        } catch(ArrayIndexOutOfBoundsException e) {
                            inFile.close();
                            throw new InvalidFormatException("Invalid Context File: Number of 'A' symbols exceeds the maximum of "+MAX_SIZE+".");
                        }

                    } else if(c == 'X') { // 'X' corresponds to a pixel which will not be used in the context, so ignore them
                        ;

                    } else if(c == '?') { // '?' corresponds to the location of the pixel currently being looked at.
                        if(qx != -1) {
                            inFile.close();
                            throw new InvalidFormatException("Invalid Context File: More than one '?' symbol found.");
                        }
                        qx = i;
                        qy = yNum;

                    } else {
                        inFile.close();
                        throw new InvalidFormatException("Invalid Context File: Found symbol '"+c+"'. Only symbols 'A', 'X', and '?' are allowed.");
                    }
                }

                yNum++;
                line = inFile.readLine();
            }
            if(qx == -1) {
                inFile.close();
                throw new InvalidFormatException("Invalid Context File: No '?' symbol found.");
            }

            // readjust the coords based on location of '?' symbol
            Coords[] temp = new Coords[arrayPos];
            for(int i=0; i < arrayPos; i++) {
                temp[i] = new Coords(coords[i].x - qx, coords[i].y - qy);
                if(Math.abs(temp[i].x) > MAX_DISTANCE || Math.abs(temp[i].y) > MAX_DISTANCE)
                    throw new InvalidFormatException("Invalid Context File: Found 'A' symbol with coordinates ("+temp[i].x+","+temp[i].y+"). Maximum allowable distance from '?' is "+MAX_DISTANCE+".");
            }
            coords = temp;
            inFile.close();

        } catch(FileNotFoundException e) {
            // file not found so use the default context
            System.out.println("Context file not found. Using default set.");
            coords = new Coords[3];
            coords[0] = new Coords(-1,-1);
            coords[1] = new Coords(0,-1);
            coords[2] = new Coords(-1,0);
        }
        
        return coords;
    }
}