import java.io.*;

/**
 * Class for Image objects.
 * 
 * @author Amanda Askew & Dane Barney
 * @version 2003.09.24
 * 
 * Latest addition: added support for PGM files
 */
public class Image {
    public static final int BITMAP = 0;
    public static final int PGM = 1;
    // add more filetypes

    public int height;
    public int width;
    public short bitDepth;
    public int type;
    public int filesize;
    public byte[] header;
    public int[][] pixels;
    public int[] colorTable; // stores the actual pixel colors as RGB values (needed for displaying images in a window)
    public int dominantColor;
    
    /**
     * Create an image object from a file
     */
    public Image(String filename, boolean keepHeader) throws IOException, InvalidFormatException {
        BitReader in = new BitReader(new DataInputStream(new FileInputStream(filename)));
        
        getType(filename);
        
        switch(type) {
            case BITMAP:
                readBitmapHeader(in, keepHeader);
                pixels = new int[width][height];
                fillBitmapPixels(in);
                break;
            case PGM:
                readPGMHeader(in, keepHeader);
                pixels = new int[width][height];
                fillRasterPixels(in);
                break;
            // add more filetypes
        }
        System.out.println("Image header size = "+header.length);
        
        in.close();

	System.out.println("height: " + height);
	System.out.println("width: " + width);
    }
    
    /**
     * Create an empty image object of specified type, reading in the header from the input stream
     */
    public Image(int type, BitReader in, boolean keepHeader) throws InvalidFormatException {
        this.type = type;
        
        switch(type) {
            case BITMAP:
                readBitmapHeader(in, keepHeader);
                pixels = new int[width][height];
                break;
            case PGM:
                readPGMHeader(in, keepHeader);
                pixels = new int[width][height];
                break;
            // add more filetypes
        }
    }
    
    /**
     * Create an exact duplicate of an image, with the option of copying pixel values
     */
    public Image(Image im, boolean copyPixels) {
        height = im.height;
        width = im.width;
        bitDepth = im.bitDepth;
        type = im.type;
        filesize = im.filesize;
        header = im.header;
        colorTable = im.colorTable;
        dominantColor = im.dominantColor;
        if(copyPixels)
            pixels = im.pixels;
        else
            pixels = new int[width][height];
    }
    
    /**
     * Writes the image to a file
     */
    public void save(String filename) throws InvalidFormatException, IOException {
        BitWriter out = new BitWriter(new DataOutputStream(new FileOutputStream(filename)));
        
        // first write header to file
        for(int i=0; i < header.length; i++)
            out.writeByte(header[i]);
        
        switch(type) {
            case BITMAP: saveBitmapPixels(out); break;
            case PGM: saveRasterPixels(out); break;
            // add more filetypes
        }
        
        out.close();
    }
    
    private void saveBitmapPixels(BitWriter out) throws InvalidFormatException, IOException {
        /* The bitmap file format requires that each line of pixels be a multiple of 32-bits, and if it
        is not, then 0's are added until a multiple of 32 bits is reached. The zeroPaddedBits variable
        keeps track of how many of these 0's are included, based on the width dimension of the bitmap,
        and will skip over these bits when reading from the file because they do not represent pixel data. */
        int zeroPaddedBits = (width * bitDepth) % 32 == 0 ? 0 : (32 - ((width * bitDepth) % 32));
        
        for(int j=height-1; j >= 0; j--) { // iterate through each row (reverse ordering)
            for(int i=0; i < width; i++) { // iterate through each pixel in this row
                switch(bitDepth) {
                    case 1: out.writeBit(pixels[i][j]); break;
                    case 4: out.writeNibble(pixels[i][j]); break;
                    case 8: out.writeByte((byte)pixels[i][j]); break;
                    case 16: out.write2Bytes((short)pixels[i][j]); break;
                    case 24: out.write3Bytes(pixels[i][j]); break;
                    //case 32: out.write4Bytes(pixels[i][j]); break;
                    default: throw new InvalidFormatException("Unsupported image bit-depth: "+bitDepth);
                }
            }
            // write all zero-padded bits in this line
            for(int i=0; i < zeroPaddedBits; i++)
                out.writeBit(0);
        }
    }

    private void saveRasterPixels(BitWriter out) throws InvalidFormatException, IOException {
        for(int j=0; j < height; j++) { // iterate through each row (top to bottom)
            for(int i=0; i < width; i++) { // iterate through each pixel in this row (left to right)
                switch(bitDepth) {
                    case 1: out.writeBit(pixels[i][j]); break;
                    case 4: out.writeNibble(pixels[i][j]); break;
                    case 8: out.writeByte((byte)pixels[i][j]); break;
                    case 16: out.write2Bytes((short)pixels[i][j]); break;
                    case 24: out.write3Bytes(pixels[i][j]); break;
                    default: throw new InvalidFormatException("Unsupported image bit-depth: "+bitDepth);
                }
            }
        }
    }

    /**
     * Reads in byte-by-byte the bitmap header data. Also extracts width, height, and
     * bit-depth information of the bitmap from the header.
     */
    private void readBitmapHeader(BitReader in, boolean keepHeader) throws InvalidFormatException {
        try {
            int pos = 0; // counter for current byte # in header
            
            // two placeholders for temporarily storing two and four bytes of data
            short current2Bytes;
            int current4Bytes;
            
            // first 2 bytes = file format code; must be 'BM' for bitmap
            short bfType = in.read2Bytes(true);
            if(bfType != 16973) /* Make sure first 2 bytes = 'BM', indicating bitmap format */ {
                in.close();
                throw new InvalidFormatException("Invalid input file. Header is not recognized.");
            }
            
            // next 4 bytes = file size in bytes
            int bfSize = in.read4Bytes(true);
            filesize = Tools.toBigEndian(bfSize);
            
            // skip next 4 bytes
            int bfReserved = in.read4Bytes(true);
            
            // next 4 bytes = byte offset from beginning of file at which the pixel data begins (the header size in bytes)
            current4Bytes = in.read4Bytes(true);
            int headerSize = Tools.toBigEndian(current4Bytes);
            
            // now write all previous bytes into the header[] array
            if(keepHeader) header = new byte[headerSize];
            if(keepHeader) put2Bytes(bfType, pos); pos+=2;
            if(keepHeader) put4Bytes(bfSize, pos); pos+=4;
            if(keepHeader) put4Bytes(bfReserved, pos); pos+=4;
            if(keepHeader) put4Bytes(current4Bytes, pos); pos+=4;
            
            // skip next 4 bytes
            current4Bytes = in.read4Bytes(true);
            if(keepHeader) put4Bytes(current4Bytes, pos); pos+=4;
            
            // next 4 bytes = width of bitmap in pixels
            current4Bytes = in.read4Bytes(true);
            width = Tools.toBigEndian(current4Bytes);
            if(keepHeader) put4Bytes(current4Bytes, pos); pos+=4;
            
            // next 4 bytes = height of bitmap in pixels
            current4Bytes = in.read4Bytes(true);
            height = Tools.toBigEndian(current4Bytes);
            if(keepHeader) put4Bytes(current4Bytes, pos); pos+=4;
            
            // skip next 2 bytes
            current2Bytes = in.read2Bytes(true);
            if(keepHeader) put2Bytes(current2Bytes, pos); pos+=2;
            
            // next 2 bytes = bit-depth of the bitmap
            current2Bytes = in.read2Bytes(true);
            bitDepth = Tools.toBigEndian(current2Bytes);
            if(keepHeader) put2Bytes(current2Bytes, pos); pos+=2;
            
            // next 4 bytes = 0 if no RLE compression
            current4Bytes = in.read4Bytes(true);
            if(current4Bytes != 0) {
                in.close();
                throw new InvalidFormatException("This bitmap is RLE compressed. Only uncompressed bitmaps allowed.");
            }
            if(keepHeader) put4Bytes(current4Bytes, pos); pos+=4;
            
            // skip next 20 bytes
            for(int i=0; i < 20; i += 4) {
                current4Bytes = in.read4Bytes(true);
                if(keepHeader) put4Bytes(current4Bytes, pos); pos+=4;
            }
            
            // now read in the color table data (if not 24-bit)
            if(bitDepth < 24) {
                colorTable = new int[(int)Math.pow(2, bitDepth)];
                
                for(int i=0; pos < headerSize; i++) {
                    int b = Tools.unsigned(in.readByte(true)); if(keepHeader) header[pos] = (byte)b; pos++;
                    int g = Tools.unsigned(in.readByte(true)); if(keepHeader) header[pos] = (byte)g; pos++;
                    int r = Tools.unsigned(in.readByte(true)); if(keepHeader) header[pos] = (byte)r; pos++;
                    byte temp = in.readByte(true); if(keepHeader) header[pos] = temp; pos++;
                    
                    int rgb = (r << 16) | (g << 8) | b;
                    colorTable[i] = rgb;
                }
            }

            assert (pos == headerSize);
        
        } catch(IOException e) {
            // if an EOFException is thrown, the header is not properly formatted
            throw new InvalidFormatException("Invalid input file. Header is not recognized.");
        }
    }
    
    /**
     * Reads in byte-by-byte the PGM header data. Also extracts width and height information of 
     * the PGM from the header.
     */
    private void readPGMHeader(BitReader in, boolean keepHeader) throws InvalidFormatException {
        bitDepth = 8;
        
        try {
            
            // temporarily store header as one long ASCII string
            String s = "";
            
            // placeholder for temporarily storing one ASCII character
            char c;
            
            // first 2 bytes = file format code; must be 'P5' for PGM
            c = (char)in.readByte(true);
            if(c != 'P') {
                in.close();
                throw new InvalidFormatException("Invalid input file. Header is not recognized.");
            }
            s += c;
            c = (char)in.readByte(true);
            if(c != '5') {
                in.close();
                throw new InvalidFormatException("Invalid input file. Header is not recognized.");
            }
            s += c;
            
            // next byte = whitespace
            s += (char)in.readByte(true);
            
            // skip any comment lines
            c = (char)in.readByte(true);
            while(c == '#') {
                s += c;
                // continue until new-line character
                c = (char)in.readByte(true);
                while(c != 10) {
                    s += c;
                    c = (char)in.readByte(true);
                }
                s += c;
                c = (char)in.readByte(true);
            }
            
            assert Character.isDigit(c); // width should be next
            
            // next few bytes = image width
            while(Character.isDigit(c)) {
                s += c;
                width = width*10 + Character.digit(c,10);
                c = (char)in.readByte(true);
            }
            s += c;
            
            // next few bytes = image height
            c = (char)in.readByte(true);
            while(Character.isDigit(c)) {
                s += c;
                height = height*10 + Character.digit(c,10);
                c = (char)in.readByte(true);
            }
            s += c;
            
            // skip any comment lines
            c = (char)in.readByte(true);
            while(c == '#') {
                s += c;
                // continue until new-line character
                c = (char)in.readByte(true);
                while(c != 10) {
                    s += c;
                    c = (char)in.readByte(true);
                }
                s += c;
                c = (char)in.readByte(true);
            }
            
            assert Character.isDigit(c); // max pixel value should be next

            // next few bytes = max pixel value
            while(Character.isDigit(c)) {
                s += c;
                c = (char)in.readByte(true);
            }
            s += c;
            
            // copy over header string to header byte array
            if(keepHeader) {
                header = new byte[s.length()];
                for(int i=0; i < header.length; i++)
                    header[i] = (byte)s.charAt(i);
            }
            
        } catch(IOException e) {
            // if an EOFException is thrown, the header is not properly formatted
            throw new InvalidFormatException("Invalid input file. Header is not recognized.");
        }
    }
    
    private void fillBitmapPixels(BitReader in) throws InvalidFormatException, IOException {
        
        // keep track of the number of occurences of each pixel (to set dominantColor)
        int possibleValues = (int)Math.pow(2, bitDepth);
        int[] counts = new int[possibleValues];
        
        /* The bitmap file format requires that each line of pixels be a multiple of 32-bits, and if it
        is not, then 0's are added until a multiple of 32 bits is reached. The zeroPaddedBits variable
        keeps track of how many of these 0's are included, based on the width dimension of the bitmap,
        and will skip over these bits when reading from the file because they do not represent pixel data. */
        int zeroPaddedBits = (width * bitDepth) % 32 == 0 ? 0 : (32 - ((width * bitDepth) % 32));
        
        for(int j=height-1; j >= 0; j--) { // iterate through each row (reverse ordering)
            for(int i=0; i < width; i++) { // iterate through each pixel in this row
                int x;
                switch(bitDepth) {
                    case 1: x = Tools.unsigned(in.readBit(true)); break;
                    case 4: x = Tools.unsigned(in.readNibble(true)); break;
                    case 8: x = Tools.unsigned(in.readByte(true)); break;
                    case 16: x = Tools.unsigned(in.read2Bytes(true)); break;
                    case 24: x = Tools.unsigned(in.read3Bytes(true)); break;
                    //case 32: x = Tools.unsigned(in.read4Bytes(true)); break;
                    default: throw new InvalidFormatException("Unsupported image bit-depth: "+bitDepth);
                }
                pixels[i][j] = x;
                counts[x]++;
            }
            // skip over all zero-padded bits in this line
            for(int i=0; i < zeroPaddedBits; i++)
                in.readBit(true);
        }
        
        // determine dominantColor
        int maxCount = 0;
        for(int i=0; i < counts.length; i++) {
            if(counts[i] > maxCount) {
                maxCount = counts[i];
                dominantColor = i;
            }
        }
    }
    
    private void fillRasterPixels(BitReader in) throws InvalidFormatException, IOException {
        
        // keep track of the number of occurences of each pixel (to set dominantColor)
        int possibleValues = (int)Math.pow(2, bitDepth);
        int[] counts = new int[possibleValues];
        
        for(int j=0; j < height; j++) { // iterate through each row (top to bottom)
            for(int i=0; i < width; i++) { // iterate through each pixel in this row (left to right)
                int x;
                switch(bitDepth) {
                    case 1: x = Tools.unsigned(in.readBit(true)); break;
                    case 4: x = Tools.unsigned(in.readNibble(true)); break;
                    case 8: x = Tools.unsigned(in.readByte(true)); break;
                    case 16: x = Tools.unsigned(in.read2Bytes(true)); break;
                    case 24: x = Tools.unsigned(in.read3Bytes(true)); break;
                    default: throw new InvalidFormatException("Unsupported image bit-depth: "+bitDepth);
                }
                pixels[i][j] = x;
                counts[x]++;
            }
        }
        
        // determine dominantColor
        int maxCount = 0;
        for(int i=0; i < counts.length; i++) {
            if(counts[i] > maxCount) {
                maxCount = counts[i];
                dominantColor = i;
            }
        }
    }
    
    public int getRGB(int pixel) {
        int rgb = pixel;
        
        switch(type) {
            case BITMAP:
                if(bitDepth != 24) rgb = colorTable[pixel]; break;
            // add more filetypes
        }
        
        return rgb;
    }
    
    private void getType(String filename) throws InvalidFormatException {
        // determine image type from file suffix
        int index = filename.lastIndexOf('.');
        if(index == -1)
            throw new InvalidFormatException("Image does not have a file extension.");
        
        String suffix = filename.substring(index+1).toLowerCase();
        if(suffix.equals("bmp") || suffix.equals("dib") || suffix.equals("rle")) type = BITMAP;
        else if(suffix.equals("pgm")) type = PGM;
        // add more filetypes
        else throw new InvalidFormatException("Unrecognized file extension.");
    }

    public String sayType() {
        switch(type) {
            case BITMAP: return "bitmap";
            case PGM: return "portable gray map";
            // add more filetypes
            default: return "unrecognized";
        }
    }

    private void put2Bytes(short val, int pos) {
        header[pos] = (byte)(val >>> 8);
        header[pos+1] = (byte)val;
    }

    private void put4Bytes(int val, int pos) {
        header[pos] = (byte)(val >>> 24);
        header[pos+1] = (byte)(val >>> 16);
        header[pos+2] = (byte)(val >>> 8);
        header[pos+3] = (byte)val;
    }
}
