/**
 * Class to store constants and static methods.
 */
public class Tools {
    public static final int DOUBLE_MANTISSA_BITS = 52; // number of bits used to store the mantissa in the 64-bit IEEE-754 standard
    public static final int FLOAT_MANTISSA_BITS = 23; // number of bits used to store the mantissa in the 32-bit IEEE-754 standard
    
    public static final byte POSITIVE = 0;
    public static final byte NEGATIVE = 1;

    public static final int BLACK = 0;
    public static final int WHITE = 1;

    public static final int INSIGNIFICANT = 0;
    public static final int SIGNIFICANT = 1;

    public static int unsigned(byte b) {
        int x = b; x <<= 24; x >>>= 24;
        return x;
    }

    public static int unsigned(short s) {
        int x = s; x <<= 16; x >>>= 16;
        return x;
    }
    
    public static int unsigned(int i) {
        int x = i; x <<= 8; x >>>= 8;
        return x;
    }
    
    public static long unsigned(long l) {
        long x = l; x <<= 32; x >>>= 32;
        return x;
    }
    
    
    /**
     * Converts a 2-byte number from little endian to big endian byte order
     * (switches the 8 most-significant bits with the 8 least-significant bits).
     * 
     * @param n the 2-byte number in little endian byte order
     * @return the 2-byte number in big endian byte order
     */
    public static short toBigEndian(short n) {
        byte b0 = (byte)n;
        byte b1 = (byte)(n >>> 8);
        int i0 = b0 << 8; i0 <<= 16; i0 >>>= 16;
        int i1 = b1; i1 <<= 24; i1 >>>= 24;
        return (short)(i0+i1);
   }

    /**
     * Converts a 4-byte number from little endian to big endian byte order
     * (reverses the ordering of the 4 bytes).
     * 
     * @param n the 4-byte number in little endian byte order
     * @return the 4-byte number in big endian byte order
     */
    public static int toBigEndian(int n) {
        byte b0 = (byte)n;
        byte b1 = (byte)(n >>> 8);
        byte b2 = (byte)(n >>> 16);
        byte b3 = (byte)(n >>> 24);
        int i0 = b0 << 24;
        int i1 = b1 << 16; i1 <<= 8; i1 >>>= 8;
        int i2 = b2 << 8; i2 <<= 16; i2 >>>= 16;
        int i3 = b3; i3 <<= 24; i3 >>>= 24;
        long l = (long)(i0+i1+i2+i3); l <<= 32; l >>>= 32;
        return (int)l;
    }
    
    /**
     * Computes and prints out the PSNR of two images (which must have the same dimensions
     * and bit-depth).
     */
    public static void printPSNR(Image im1, Image im2) {
        assert (im1.height == im2.height && im1.width == im2.width && im1.bitDepth == im2.bitDepth);
        
        double mse = 0;
        for(int j=0; j < im1.height; j++)
            for(int i=0; i < im1.width; i++){
                mse += Math.pow(im1.pixels[i][j] - im2.pixels[i][j], 2);
	    }
        mse /= (double)(im1.width * im1.height);
        System.out.println("mse = "+mse);
        int maxVal = (int)Math.pow(2, im1.bitDepth) - 1;
        double x = Math.pow(maxVal, 2) / mse;
        double psnr = 10.0 * logbase10(x);
        System.out.println("PSNR = "+psnr);
    }
    
    /**
     * Returns the base-10 logarithm of a number
     */
    public static double logbase10(double x) {
        return Math.log(x) / Math.log(10);
    }
}
